@@ -60,6 +60,9 @@ GType wnck_tasklist_get_type (void) G_GNUC_CONST;
 
 GtkWidget *wnck_tasklist_new (WnckScreen *screen);
 
+const int *wnck_tasklist_get_size_hint_list (WnckTasklist  *tasklist,
+					      int           *n_elements);
+
 void wnck_tasklist_set_grouping (WnckTasklist             *tasklist,
 				 WnckTasklistGroupingType  grouping);
 void wnck_tasklist_set_switch_workspace_on_unminimize (WnckTasklist  *tasklist,
@@ -68,6 +71,11 @@ void wnck_tasklist_set_grouping_limit (WnckTasklist *tasklist,
 				       gint          limit);
 void wnck_tasklist_set_include_all_workspaces (WnckTasklist *tasklist,
 					       gboolean      include_all_workspaces);
+void wnck_tasklist_set_minimum_width (WnckTasklist *tasklist, gint size);
+gint wnck_tasklist_get_minimum_width (WnckTasklist *tasklist);
+void wnck_tasklist_set_minimum_height (WnckTasklist *tasklist, gint size);
+gint wnck_tasklist_get_minimum_height (WnckTasklist *tasklist);
+    
 G_END_DECLS
 
 #endif /* WNCK_TASKLIST_H */
