@@ -22,6 +22,7 @@
 #include <config.h>
 #include "xutils.h"
 #include <string.h>
+#include <stdio.h>
 #include "screen.h"
 #include "window.h"
 #include "private.h"
@@ -698,6 +699,12 @@ filter_func (GdkXEvent  *gdkxevent,
           _wnck_window_process_configure_notify (window, xevent);
       }
       break;
+
+    case SelectionClear:
+      {
+        _wnck_desktop_layout_manager_process_event (xevent);
+      }
+      break;
     }
   
   return GDK_FILTER_CONTINUE;
@@ -763,7 +770,8 @@ _wnck_deiconify (Window xwindow)
 }
 
 void
-_wnck_close (Window xwindow)
+_wnck_close (Screen *screen,
+	     Window  xwindow)
 {
   XEvent xev;
   
@@ -779,14 +787,15 @@ _wnck_close (Window xwindow)
   xev.xclient.data.l[2] = 0;
 
   XSendEvent (gdk_display,
-              gdk_x11_get_default_root_xwindow (),
+              RootWindowOfScreen (screen),
               False,
 	      SubstructureRedirectMask | SubstructureNotifyMask,
 	      &xev); 
 }
 
 void
-_wnck_change_state (Window   xwindow,
+_wnck_change_state (Screen  *screen,
+		    Window   xwindow,
                     gboolean add,
                     Atom     state1,
                     Atom     state2)
@@ -809,15 +818,16 @@ _wnck_change_state (Window   xwindow,
   xev.xclient.data.l[2] = state2;
 
   XSendEvent (gdk_display,
-              gdk_x11_get_default_root_xwindow (),
+	      RootWindowOfScreen (screen),
               False,
 	      SubstructureRedirectMask | SubstructureNotifyMask,
 	      &xev);
 }
 
 void
-_wnck_change_workspace (Window xwindow,
-                        int    new_space)
+_wnck_change_workspace (Screen     *screen,
+			Window      xwindow,
+                        int         new_space)
 {
   XEvent xev;
   
@@ -833,14 +843,15 @@ _wnck_change_workspace (Window xwindow,
   xev.xclient.data.l[2] = 0;
 
   XSendEvent (gdk_display,
-              gdk_x11_get_default_root_xwindow (),
+	      RootWindowOfScreen (screen),
               False,
 	      SubstructureRedirectMask | SubstructureNotifyMask,
 	      &xev);
 }
 
 void
-_wnck_activate (Window xwindow)
+_wnck_activate (Screen *screen,
+		Window  xwindow)
 {
   XEvent xev;
   
@@ -856,14 +867,15 @@ _wnck_activate (Window xwindow)
   xev.xclient.data.l[2] = 0;
 
   XSendEvent (gdk_display,
-              gdk_x11_get_default_root_xwindow (),
+	      RootWindowOfScreen (screen),
               False,
 	      SubstructureRedirectMask | SubstructureNotifyMask,
 	      &xev); 
 }
 
 void
-_wnck_activate_workspace (int new_active_space)
+_wnck_activate_workspace (Screen *screen,
+			  int     new_active_space)
 {
   XEvent xev;
   
@@ -871,7 +883,7 @@ _wnck_activate_workspace (int new_active_space)
   xev.xclient.serial = 0;
   xev.xclient.send_event = True;
   xev.xclient.display = gdk_display;
-  xev.xclient.window = gdk_x11_get_default_root_xwindow ();
+  xev.xclient.window = RootWindowOfScreen (screen);
   xev.xclient.message_type = _wnck_atom_get ("_NET_CURRENT_DESKTOP");
   xev.xclient.format = 32;
   xev.xclient.data.l[0] = new_active_space;
@@ -879,7 +891,7 @@ _wnck_activate_workspace (int new_active_space)
   xev.xclient.data.l[2] = 0;
 
   XSendEvent (gdk_display,
-              gdk_x11_get_default_root_xwindow (),
+	      RootWindowOfScreen (screen),
               False,
 	      SubstructureRedirectMask | SubstructureNotifyMask,
 	      &xev);
@@ -2000,11 +2012,12 @@ _wnck_get_fallback_icons (GdkPixbuf **iconp,
 
 
 void
-_wnck_get_window_geometry (Window xwindow,
-                           int   *xp,
-                           int   *yp,
-                           int   *widthp,
-                           int   *heightp)
+_wnck_get_window_geometry (Screen *screen,
+			   Window  xwindow,
+                           int    *xp,
+                           int    *yp,
+                           int    *widthp,
+                           int    *heightp)
 {
   int x, y, width, height, bw, depth;
   Window root_window;
@@ -2021,7 +2034,7 @@ _wnck_get_window_geometry (Window xwindow,
   
   _wnck_error_trap_pop ();
 
-  _wnck_get_window_position (xwindow, xp, yp);
+  _wnck_get_window_position (screen, xwindow, xp, yp);
 
   if (widthp)
     *widthp = width;
@@ -2030,9 +2043,10 @@ _wnck_get_window_geometry (Window xwindow,
 }
 
 void
-_wnck_get_window_position (Window xwindow,
-                           int   *xp,
-                           int   *yp)
+_wnck_get_window_position (Screen *screen,
+			   Window  xwindow,
+                           int    *xp,
+                           int    *yp)
 {
   int x, y;
   Window child;
@@ -2043,7 +2057,7 @@ _wnck_get_window_position (Window xwindow,
   _wnck_error_trap_push ();
   XTranslateCoordinates (gdk_display,
                          xwindow,
-                         gdk_x11_get_default_root_xwindow (),
+			 RootWindowOfScreen (screen),
                          0, 0,
                          &x, &y, &child);
   _wnck_error_trap_pop ();
@@ -2096,3 +2110,239 @@ _wnck_set_dock_type_hint (Window xwindow)
 
   _wnck_error_trap_pop ();
 }
+
+#define _NET_WM_ORIENTATION_HORZ 0
+#define _NET_WM_ORIENTATION_VERT 1
+
+void
+_wnck_set_desktop_layout (Screen *xscreen,
+                          int     rows,
+                          int     columns)
+{
+  gulong data[3];
+
+  data[0] = _NET_WM_ORIENTATION_HORZ;
+  data[1] = rows;
+  data[2] = columns;
+  
+  _wnck_error_trap_push ();
+
+  XChangeProperty (gdk_display,
+                   RootWindowOfScreen (xscreen),
+		   _wnck_atom_get ("_NET_DESKTOP_LAYOUT"),
+		   XA_CARDINAL, 32, PropModeReplace,
+		   (guchar *)&data, 3);
+
+  _wnck_error_trap_pop ();
+}
+
+typedef struct 
+{
+  Window window;
+  Atom timestamp_prop_atom;
+} TimeStampInfo;
+
+static Bool
+timestamp_predicate (Display *display,
+		     XEvent  *xevent,
+		     XPointer arg)
+{
+  TimeStampInfo *info = (TimeStampInfo *)arg;
+
+  if (xevent->type == PropertyNotify &&
+      xevent->xproperty.window == info->window &&
+      xevent->xproperty.atom == info->timestamp_prop_atom)
+    return True;
+
+  return False;
+}
+
+/**
+ * get_server_time:
+ * @display: display from which to get the time
+ * @window: a #Window, used for communication with the server.
+ *          The window must have PropertyChangeMask in its
+ *          events mask or a hang will result.
+ * 
+ * Routine to get the current X server time stamp. 
+ * 
+ * Return value: the time stamp.
+ **/
+static Time
+get_server_time (Display *display,
+		 Window   window)
+{
+  unsigned char c = 'a';
+  XEvent xevent;
+  TimeStampInfo info;
+
+  info.timestamp_prop_atom = XInternAtom  (display, "_TIMESTAMP_PROP", False);
+  info.window = window;
+
+  XChangeProperty (display, window,
+		   info.timestamp_prop_atom, info.timestamp_prop_atom,
+		   8, PropModeReplace, &c, 1);
+
+  XIfEvent (display, &xevent,
+	    timestamp_predicate, (XPointer)&info);
+
+  return xevent.xproperty.time;
+}
+
+typedef struct
+{
+  int screen_number;
+  int token;
+  Window window;
+  Atom selection_atom;
+  Atom manager_atom;
+} LayoutManager;
+
+static GSList *layout_managers = NULL;
+static int next_token = 1;
+
+int
+_wnck_try_desktop_layout_manager (Screen *xscreen,
+                                  int     current_token)
+{
+  GSList *tmp;
+  int number;
+  Time timestamp;
+  XClientMessageEvent xev;  
+  char buffer[256];
+  LayoutManager *lm;
+  
+  number = XScreenNumberOfScreen (xscreen);
+  
+  tmp = layout_managers;
+  while (tmp != NULL)
+    {
+      lm = tmp->data;
+
+      if (number == lm->screen_number)
+        {
+          if (current_token == lm->token)
+            return current_token; /* we still have the selection */
+          else
+            return WNCK_NO_MANAGER_TOKEN; /* someone else has it */
+        }
+      
+      tmp = tmp->next;
+    }
+  
+  /* No one in-process has the selection at the moment */
+
+  lm = g_new0 (LayoutManager, 1);
+
+  lm->screen_number = number;
+  lm->token = next_token;
+  ++next_token;
+
+  sprintf (buffer, "_NET_DESKTOP_LAYOUT_S%d", number);
+  lm->selection_atom = XInternAtom (gdk_display, buffer, False);
+  lm->manager_atom = XInternAtom (gdk_display, "MANAGER", False);
+
+  lm->window = XCreateSimpleWindow (gdk_display,
+                                    RootWindowOfScreen (xscreen),
+                                    0, 0, 10, 10, 0,
+                                    WhitePixel (gdk_display, number),
+                                    WhitePixel (gdk_display, number));
+
+  XSelectInput (gdk_display, lm->window, PropertyChangeMask);
+  timestamp = get_server_time (gdk_display, lm->window);
+
+  XSetSelectionOwner (gdk_display, lm->selection_atom,
+		      lm->window, timestamp);
+
+  /* Check to see if we managed to claim the selection. */
+
+  if (XGetSelectionOwner (gdk_display, lm->selection_atom) !=
+      lm->window)
+    {
+      g_free (lm);
+      return WNCK_NO_MANAGER_TOKEN;
+    }
+  
+  xev.type = ClientMessage;
+  xev.window = RootWindow (gdk_display, number);
+  xev.message_type = lm->manager_atom;
+  xev.format = 32;
+  xev.data.l[0] = timestamp;
+  xev.data.l[1] = lm->selection_atom;
+  xev.data.l[2] = lm->window;
+  xev.data.l[3] = 0;	/* manager specific data */
+  xev.data.l[4] = 0;	/* manager specific data */
+  
+  XSendEvent (gdk_display, RootWindow (gdk_display, number),
+              False, StructureNotifyMask, (XEvent *)&xev);
+
+  layout_managers = g_slist_prepend (layout_managers,
+                                     lm);
+  
+  return lm->token;
+}
+
+void
+_wnck_release_desktop_layout_manager (Screen *xscreen,
+                                      int     current_token)
+{
+  GSList *tmp;
+  int number;
+  LayoutManager *lm;
+  
+  number = XScreenNumberOfScreen (xscreen);
+  
+  tmp = layout_managers;
+  while (tmp != NULL)
+    {
+      lm = tmp->data;
+
+      if (number == lm->screen_number)
+        {
+          if (current_token == lm->token)
+            {
+              XDestroyWindow (gdk_display,
+                              lm->window);
+
+              g_free (lm);
+              layout_managers = g_slist_remove (layout_managers, lm);
+              
+              return;
+            }
+        }
+      
+      tmp = tmp->next;
+    }
+}
+
+gboolean
+_wnck_desktop_layout_manager_process_event (XEvent *xev)
+{
+  GSList *tmp;
+  LayoutManager *lm;
+
+  if (xev->type != SelectionClear)
+    return FALSE;
+  
+  tmp = layout_managers;
+  while (tmp != NULL)
+    {
+      lm = tmp->data;
+
+      if (xev->xany.window == lm->window &&
+          xev->xselectionclear.selection == lm->selection_atom)
+        {
+          XDestroyWindow (gdk_display,
+                          lm->window);
+          
+          g_free (lm);
+          layout_managers = g_slist_remove (layout_managers, lm);
+          
+          return TRUE;
+        }
+      
+      tmp = tmp->next;
+    }
+
+  return FALSE;
+}
